! module lorenz1.f90
! Original program file, as used in book 
! "Mathematics of Weather" (Steppeler et al.)
! 03-2022 embellished by EH
!         Evaluation of command line added for the basic three parameters
!         OK: works also in script, command line params are evaluated
! 10.03.2022 4th command line parameter added: ntotal
! 12.03.2022 Writes final amplitudes into file LorenzOutFields.txt
!
! ------------------------------------------------------------------------
PROGRAM plots3
  REAL              :: c        ! basic param: circulation
  REAL              :: fl, fm   ! basic param: temperatures
  REAL              :: dt
  INTEGER           :: ntotal   ! basic param: no. of steps
  INTEGER           :: noArgs;
  character(len=32) :: arg
  !
  !  Initial setting for the values the user has to give on the CLI
  c  = 12.3
  fl = 8.5
  fm = 44.
  !
  dt     = .1
  ntotal = 10000
  !
  ! c: amplitude defining the amount of velocity, fl and fm: 
  ! amplitudes defining the temperature
  ! dt: time step, ntotal: the total number of prognostic steps done, 
  ! ntotal*dt is the time 
  ! the time of climate prediction goes from 0 to ntotal*dt
  !
  ! Evaluate the command line arguments (3 args. needed)
  noArgs = command_argument_count()
  if (noArgs .GE. 1) THEN
    call get_command_argument(1, arg)
    READ(arg,*) fl
  ENDIF
  if (noArgs .GE. 2) THEN
    call get_command_argument(2, arg)
    READ(arg,*) fm
  ENDIF
  if (noArgs .GE. 3) THEN
    call get_command_argument(3, arg)
    READ(arg,*) c
  ENDIF
  if (noArgs .GE. 4) THEN
    call get_command_argument(4, arg)
    READ(arg,*) ntotal
  ENDIF
  !
  ! print *, 'initial: c,fl,fm:', c,fl,fm
  !
  CALL OPNGKS  ! Open GKS (or replacement routines)
  !
  CALL plot_3(dt,ntotal,c,fl,fm)
  !
  CALL CLSGKS  ! Close GKS (or replacement routines)
END PROGRAM plots3 

! -------------------------------------------------------------------
! Writes thre REAL output values in three lines into a file
! This file be be later read by a CGI C++ program
SUBROUTINE WriteOutFields(val1, val2, val3)
  REAL    :: val1, val2,val3
  INTEGER :: io_error
  !
  OPEN(unit=31,                    &
       file='lorenzOutFields.txt', &
       status='replace',           &
       action='write',             &
       iostat=io_error)
  if (io_error .eq. 0) THEN
    WRITE (31, '(f0.2)') val1
    WRITE (31, '(f0.2)') val2
    WRITE (31, '(f0.2)') val3
    !
    CLOSE(unit=31) 
  END IF
  RETURN
END SUBROUTINE WriteOutFields

! -------------------------------------------------------------------
SUBROUTINE plot_3(dt, ntotal, & 
                  c, fl, fm)
  REAL  :: dt
  REAL  :: c, fl, fm 
  !
  CALL SET(0. , 1. , 0. , 1. , -5. , 5. , -5. , 5. , 1)
  !
  dt  = .01
  sig = 10.
  b   = 8./3.
  r   = 28
  do it = 1, ntotal
    !do it = 1,1000000
    c1  = c+dt*(sig*(fl-c))
    fl1 = fl+dt*(r*c-fl-c*fm)
    fm1 = fm+dt*(c*fl-b*fm)
    CALL LINE(fm*4. / 50.,  c * 4. / 20. , fm1 * 4. / 50. , c1 * 4. / 20.)
    c  = c1
    fl = fl1
    fm = fm1
    if (abs(c).ge.1000.) print*,'approximation error'
    if (abs(c).ge.1000.) stop
  enddo
  ! print *, c,fl,fm,'c,fl,fm'! Dies sind die Amplitudenwerte 
                            ! am Ende der Vorhersage, die
                            ! als Anfangswerte fuer eine neue 
                            ! Vorhersage verwendet werden koennen. 
  !   CALL LINE ( -4. ,  0. , 4. , 0. )
  !   CALL LINE ( -4. , -4. , 4. , 4. )
  CALL WriteOutFields(fl, fm, c)
  !
  CALL FRAME
END SUBROUTINE plot_3
