// module metereo2.cpp
// Fuer Projekt MoW (J. Steppeler/Jinxi Li)
// CGI Programm zum Ablaufen in einem WEB Server
// Ist im Kern ein einfaches C-Programm, kein C++ Programm!
// Dr. E.Huckert 09-2021, 03-2022
//
// Version : 10-2008, 06-2016, 11-2019, 09-2021, 03-2022
// 05.03.2022 version metereo2 prepared for use with Lorenz chaos model
//            derived from metereo1.cpp
// 10.03.2022 4th parameter (feld4=STEPS) added
//
// Unter Linux:
//   compile: g++ -o metereo2 metereo2.cpp hu_http.cpp
//   Executable ablegen in /usr/lib/cgi-bin
//   Aufruf: z.B. huckert9/cgi-bin/metereo2?1 oder 

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/timeb.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/stat.h>
#else
#include <windows.h>
#include <io.h>
#endif

#include "hu_http.h"
#include "inifile.h"

#define LBUF 512

// globale Variablen
char proto_fn[128] = "./metereo2.txt";   
int  proto_error   = 0;
FILE *ofd          = NULL;  // used for logging - nothing else
char logo_ref[]    = "<IMG SRC=/icons/kfw1.gif ALIGN=RIGHT BORDER=1><BR><BR>";
char doc_root[128] = "/usr/var/html";  // ersetzt durch DOCUMENT_ROOT
char host_name[128]      = "huckert9";   // ersetzt durch HTTP_HOST
char client_addr[128]    = "";        // == Inhalt von REMOTE_ADDR
char request_method[128] = "";
char query_string[2048];
char *parbuf             = NULL;
char zeit1[20];
char zeit2[20];
int  logging              = 1;
HU_HTTP *pGlobalHttp      = NULL;
IniFile *pIni             = NULL;
const char *pIniFileName  = "./metereo2.ini";
char  templatePageFn[256] = "./answerLorenz.html";  //may be overwritten!
const char *pProjectTitle = "MoW project J.Steppeler";
const char *pLorenzOutFieldsFn = "./lorenzOutFields.txt";

// --------------------------------------------------------
// liefert Zeit im Format SS:MM:SS:mmm 
// The returned pointer must be freed externally!
char *getTime()
{
  time_t   ltime;
  struct tm *newtime;
  struct timeb tb;
  char *retbuf;
  time(&ltime);
  newtime = localtime(&ltime);
  ::ftime(&tb);
  retbuf = new char[16];
  if (retbuf != NULL)
  {
    sprintf(retbuf,"%02d:%02d:%02d.%03d",
          (int)newtime->tm_hour,
          (int)newtime->tm_min,
          (int)newtime->tm_sec,
          (int)tb.millitm);
  }
  return retbuf;
}   // end get_time()

// ---------------------------------------------------------------------
// liefert Datum im Format YYYYMMDD 
// Returns: an allocated(!) C-string containing the date
//          The returned string must be freed externally
char *getDateYYYYMMDD()
{
  time_t   ltime;
  struct tm *newtime;
  short  offset;
  char   *pBuf = NULL;
  //
  time(&ltime);
  newtime = localtime(&ltime);
  offset  = 1900;
  pBuf = new char[64];
  ::sprintf(pBuf,
            "%04u%02u%02u",
            (int)newtime->tm_year + offset,
            (int)newtime->tm_mon+1,
            (int)newtime->tm_mday);
  return pBuf;
}   // end getDateYYYYMMDD()


// -------------------------------------------------------------------
// liefert Datum im Format TT.MM.JJJJ
char *get_german_date()
{
static char dateBuf[64];
long   ltime;
struct tm *newtime;
short  offset;
  time(&ltime);
  newtime = localtime(&ltime);
  offset  = 1900;
  sprintf(dateBuf,"%02d.%02d.%04d",
          (int)newtime->tm_mday,
          (int)newtime->tm_mon+1,
          (int)newtime->tm_year + offset);
  return dateBuf;
}   // end get_german_date()


// -------------------------------------------------------------------
// Konvertiert Format JJJJMMDD in TT.MM.JJJJ
char *make_german_date(char *d)
{
static char retbuf[20];
  retbuf[0] = 0;
  strncpy(retbuf, d+6, 2);
  retbuf[2] = '.';
  strncpy(&retbuf[3], d+4, 2);
  retbuf[5] = '.';
  strncpy(&retbuf[6], d, 4);
  retbuf[10] = 0;
  return retbuf;
}   // end make_german_date()

// -------------------------------------------------------------------
// liefert Zeit im Format HH:MM:SS.mmm 
char *get_time()
{
long   ltime;
struct tm *newtime;
short  offset;
struct timeb tb;
static char dest[128];
  time(&ltime);
  newtime = localtime(&ltime);
  ftime(&tb);
  sprintf(dest,"%02d:%02d:%02d.%03d",
          (int)newtime->tm_hour,
          (int)newtime->tm_min,
          (int)newtime->tm_sec,
          (int)tb.millitm);
  return dest;
}   // end get_time()

// -------------------------------------------------------------------
// read and update a file conatining the transaction ID
// TODO: must be protocted via mutex!
// Returns: < 0L if error
long  getNextRunID()
{
  FILE *fd   = NULL;
  long runId = 0L;
  char buf[64];
  static const char *runIdFn = "./runid.txt";
  //
  fd = ::fopen(runIdFn, "r");
  if (fd == NULL)
  {
    runId = -1L;
    goto zurueck;
  } 
  if (::fgets(buf, sizeof(buf), fd) == NULL)
  {
    runId = -2L;
    goto zurueck;
  }
  runId = ::atol(buf);
  ::fclose(fd);
  //
  fd = ::fopen(runIdFn, "w");
  if (fd == NULL)
  {
    runId = -3L;
    goto zurueck;
  }
  // write back the next run ID
  ::sprintf(buf, "%ld", runId + 1);
  ::fprintf(fd, "%s", buf);
  //
  zurueck:
  if (fd != NULL)
    ::fclose(fd);
  return runId;
}   // getNextRunID()

// -------------------------------------------------------------------
// strips off LF from the given C-string
void no_lf(char *p)
{
  int n;
  //
  for (n=0; p[n] != 0; n++)
    if (p[n] == '\n')
       p[n] = ' ';
}   // end no_lf

// ----------------------------------------------------------------
// corresponds to getBasename() under UNIX
// Allocates and returns a new buffer!
// Ex.: returns "file" "for c:\xyz\file.txt"
char *getBasename(const char *path)
{
  int  n, ll, p, start;
  char *retbuf;
  //
  ll     = strlen(path);
  retbuf = new char[ll + 1];
  start = 0;
  p     = -1;
  for (n=ll - 1; n >= 0; n--)
  {
    if (path[n] == '.')  p = n;
    if ((path[n] == '\\') || (path[n] == '/')) 
    {
      start = n + 1;
      break;
    }
  }
  ::strcpy(retbuf, &path[start]);
  if (p > 0)
    retbuf[p - start] = 0;
  return retbuf;
}   // end getBasename()

// ----------------------------------------------------------------
// extrahiert einen Feldinhalt aus dem query string qs
// Ergebnis: in Variable content
int get_content(const char *qs,         // 
                const char *kat,        // Feldname, z.B. "ZSTI="
                char *content,          // Ergebnispuffer
                int lcontent)
{
  int  i, m, n;
  char *caddr;
  char ch;
  if (logging > 0)
  {
    ::fprintf(ofd,"get_content():kat=%s qs=%s\n", kat, qs);
  }
  n = m = 0;
  content[0] = 0;
  caddr = ::strstr((char *)qs, (char *)kat);
  if (caddr == NULL)
  {
    if (logging > 0)
     ::fprintf(ofd, "   kat nicht gefunden\n");
    return 0;
  }
  //
  // uebergebene Argumente bereinigt zurueckgegeben
  caddr += strlen(kat);
  while (*caddr != 0)
  {
    if (*caddr == '&')
      break;
    ch = *caddr;
    if (ch == '+') ch = ' ';
    if (ch == '%')
    {
      sscanf(caddr+1, "%02x", &i);
      ch = i & 0xff;
      caddr++;
      caddr++;
    }
    if (m < (lcontent - 2))
      content[m++] = ch;
    caddr++;
  }
  content[m] = 0;
  if (logging > 0)
  {
    fprintf(ofd,"   zurueck: %s\n", content);
  }
  return 0;
}   /* end get_content() */

// -------------------------------------------------------------------
void html_prologue(const char *title)
{
  pGlobalHttp->send_header("Content-type:", "text/html\n");
  printf("<HTML>\n");
  if ((title != NULL) && (::strlen(title) > 0))
    printf("<TITLE>%s</TITLE>\n", title);
  printf("<BODY BGCOLOR=\"LIGHTGREEN\">\n");
  printf("<FONT FACE=ARIAL>\n");
}   // end html_prologue()

// -------------------------------------------------------------------
void html_epilogue()
{
  printf("</FONT>\n");
  printf("</BODY>\n");
  printf("</HTML>\n");
}   // end html_epilogue()

// -------------------------------------------------------------------
// send back a HTML document containing an error message
int send_error_document(const char *pTitle,
                        const char *msg)
{
  html_prologue(pTitle);
  printf("<H3>%s</H3><P>\n",  msg);
  html_epilogue();
  return 0;
}   // end send_error_document()

// --------------------------------------------------------------------
// Important: assumes that html_prologue() has been called!
int html_send(const char *pMsg,
             const char *pArg)
{
  printf("<H3>%s%s</H3>\n", pMsg, pArg);
  return 0;
}   // end html_send()

// --------------------------------------------------------------------------
// Evaluierung der wichtigsten Umgebungsvariablen des HTTP Requests
// Puts the contents into fixed (static) strings
int get_env_vars()
{
  char *caddr;
  //
  caddr = ::getenv("REQUEST_METHOD");
  if (caddr != NULL)
  {
    if (ofd != NULL)
      fprintf(ofd, "REQUEST_METHOD=%s\n", caddr);
    ::strcpy(request_method, caddr);
  }
  caddr = ::getenv("HTTP_HOST");
  if (caddr != NULL)
  {
    // Dies ist die eigene (Server) Adresse
    if (ofd != NULL) fprintf(ofd, "HTTP_HOST=%s\n", caddr);
    if (strlen(caddr) > 0)
    {
      ::strcpy(host_name, "http://");
      strcat(host_name, caddr);
    }
  }
  caddr = ::getenv("REMOTE_HOST");
  if (caddr != NULL)
  {
    if (ofd != NULL) 
      fprintf(ofd, "REMOTE_HOST=%s\n", caddr);
  }
  caddr = ::getenv("REMOTE_ADDR");
  if (caddr != NULL)
  {
    if (ofd != NULL)
      fprintf(ofd, "REMOTE_ADDR=%s\n", caddr);
    ::strcpy(client_addr, caddr);
  }
  caddr = ::getenv("DOCUMENT_ROOT");
  if (caddr != NULL)
  {
    if (ofd != NULL) fprintf(ofd, "DOCUMENT_ROOT=%s\n", caddr);
    if (strlen(caddr) > 0)
    {
      ::strcpy(doc_root, caddr);
      if (ofd != NULL) 
        fprintf(ofd, "DOCUMENT_ROOT=%s\n", caddr);
    }
  }
  caddr = ::getenv("CONTENT_TYPE");
  if (caddr != NULL)
  {
    if (ofd != NULL)
      fprintf(ofd, "CONTENT_TYPE=%s\n", caddr);
  }

  query_string[0] = 0;
  caddr = ::getenv("QUERY_STRING");
  if (caddr != NULL)
  {
    if (::strlen(caddr) < sizeof(query_string))
      ::strcpy(query_string, caddr);
    else
      ::strcpy(query_string, "");
    if (ofd != NULL)
      ::fprintf(ofd, "query_string=%s\n", query_string);
  }
  return 0;
}   // end get_env_vars()

// ------------------------------------------------------------------
// Evaluates the content given in the HTML request after "CONTENT_LENTH"
// Returns: an allocated(!) C string containing the contents
char *get_query_params()
{
  char *parbuf = NULL;
  char *caddr;
  int ll;
  /* Get the env. var. defining the length of the data on stdin */
  caddr = ::getenv("CONTENT_LENGTH");
  if (caddr != NULL)
  {
    ll = ::atoi(caddr);
    if (ofd != NULL)
      ::fprintf(ofd, "CONTENT_LENGTH=%d\n", ll);
    parbuf = new char[ll+1];
    if (parbuf == NULL)
    {
      send_error_document(::pProjectTitle,
                          "ERROR: internal memory allocation");
      return NULL;
    }
    //
    // read the parameters from stdin into buffer parbuf 
    int n;
    n = ::read(0, parbuf, ll);
    parbuf[ll] = 0;
    if (ofd != NULL) ::fprintf(ofd, "content=\n");
    if (ofd != NULL) ::fprintf(ofd, "%s\n", parbuf);  
  }
  else
  {
    parbuf = new char[1];
    parbuf[0] = 0;
  }
  return parbuf;
}   // end get_query_params()

// ------------------------------------------------------------------------
bool isNumber(const char *ps)
{
  bool ret = true;
  //
  if (ps == NULL)
  {
    ret = false;
    goto zurueck;
  }
  if (::strlen(ps) == 0)
  {
    // empty field: not allowed
    ret = false;
    goto zurueck;
  }
  for (int n=0; ps[n] != 0; n++)
  {
    if (::strchr("0123456789.- ", ps[n]) == NULL)
    {
      ret = false;
      break;
    }
  }   // end for ...
  //
  zurueck:
  return ret;
}   // end isNumber()

// ----------------------------------------------------------------------
// TODO: eliminate this - not needed
// Reads script in the template nScriptPath and replaces the three given
// parameters in this script. 
// The parameters to be replaced must be named "$PARAMx".
// An output script outScriptPath is then written.
// The template inScriptPath is left unchanged
int replaceInScript(const char *inScriptPath, 
                    const char *outScriptPath,
                    const char *p1,
                    const char *p2,
                    const char *p3)
{
  int ret = 0;
  FILE *scriptIfd = NULL;
  FILE *scriptOfd = NULL;
  char lineBuf[256];
  //
  zurueck:
  if (scriptIfd != NULL)
    ::fclose(scriptIfd);
  if (scriptOfd != NULL)
    ::fclose(scriptOfd);
  return ret;
}   // end replaceInScript

// ----------------------------------------------------------
// Fills fields pfx from a file write by prorgam lorenz1EH
// Returns the number of fields successfully read
//         < 0 upon error
int readOutputFields(char *pF1,
                     char *pF2,
                     char *pF3,
                     int fieldLen)
{
  int ret = 0;
  FILE *ifd = NULL;
  char *pRbuf = NULL;
  //
  if (pF1 == NULL)
  {
    ret = -1;
    goto zurueck;
  }
  ifd = ::fopen(::pLorenzOutFieldsFn, "r");
  if (ifd == NULL)
  {
    ret = -2;
    goto zurueck;
  }
  pRbuf = new char[fieldLen];
  ::strcpy(pF1, "");
  ::strcpy(pF2, "");
  ::strcpy(pF3, "");
  if (::fgets(pF1, fieldLen, ifd) == NULL)
  {
    ret = -3;
    goto zurueck;
  }
  if ((ofd != NULL) && (::logging > 0))
    ::fprintf(ofd, "readOutputFields() f1=%s\n", pF1);
  ret++;
  if (::fgets(pF2, fieldLen, ifd) == NULL)
  {
    ret = -4;
    goto zurueck;
  }
  if ((ofd != NULL) && (::logging > 0))
    ::fprintf(ofd, "readOutputFields() f2=%s\n", pF2);
  ret++;
  if (::fgets(pF3, fieldLen, ifd) == NULL)
  {
    ret = -5;
    goto zurueck;
  }
  if ((ofd != NULL) && (::logging > 0))
    ::fprintf(ofd, "readOutputFields() f3=%s\n", pF3);
  ret++;
  //
  zurueck:
  if (ifd != NULL)
    ::fclose(ifd);
  if (pRbuf != NULL)
    delete pRbuf;
  return ret;
}   // end readOutputFields()

// ------------------------------------------------------------------------
// Kernel of the CGI program: controls the generation of a JPG file
// showing the results of a Lorenz model calculation
// Uses shell script runLorenz for the essential steps in the logic 
int lorenzModel()
{
  int iRet     = 0;
  int localRet = 0;
  long runId   = 0L;
  char xBuf[128];
  char feld1[64];  // Input
  char feld2[64];  // Input
  char feld3[64];  // Input
  char feld4[64];  // Input
  char feld5[64] = "";  // Output
  char feld6[64] = "";  // Output
  char feld7[64] = "";  // output
  char result[64];
  char *pDateStr = NULL;
  char *pTimeStr1  = NULL;
  char *pTimeStr2 = NULL;
  bool wff1, wff2, wff3, wff4;
  char *caddr = NULL;
  char sysCmd[256];
  // Liste der Felder im Antwort-Template 
  // Die Liste vewendet "|" als Separator 
  static char itemList[LBUF]  =  
     "feld1|feld2|feld3|feld4|feld5|feld6|feld7|result|date|time1|time2|runid|url1|";
  // Liste der Werte die die in itemList angegebenen Felder ersetzen
  // Muss zur Laufzeit in der richtigen Reihenfolge aufgebaut werden
  // Die Liste vewendet "|" als Separator 
  static char valueList[LBUF] = "";
  //
  if ((ofd != NULL) && (::logging > 0))
  {
    ::fprintf(ofd, "lorenzModel() begin\n");
  }
  pTimeStr1 = getTime();;  // begin of execution
  //
  // get the input fields (as entered by the user) from ::parbuf
  get_content(::parbuf, "TEMP1=", feld1, sizeof(feld1));
  get_content(::parbuf, "TEMP2=", feld2, sizeof(feld2));
  get_content(::parbuf, "CIRC=",  feld3, sizeof(feld3));
  get_content(::parbuf, "STEPS=", feld4, sizeof(feld4));
  //
  // plausibility checks
  wff1 = isNumber(feld1);
  wff2 = isNumber(feld2);
  wff3 = isNumber(feld3);
  wff4 = isNumber(feld4);
  if (! (wff1 & wff2 & wff3 & wff4))
  {
    ::pGlobalHttp->send_error_document(::pProjectTitle,
                  "ERROR: input field(s) empty or wrong format");
    iRet =  - 1;
    goto zurueck;
  }
  //
  ::strcpy(sysCmd, "./runLorenz ");
  ::strcat(sysCmd, feld1);
  ::strcat(sysCmd, " ");
  ::strcat(sysCmd, feld2);
  ::strcat(sysCmd, " ");
  ::strcat(sysCmd, feld3);
  ::strcat(sysCmd, " ");
  ::strcat(sysCmd, feld4);
  ::strcat(sysCmd, " ");
  //
  // field feld5 .. feld7 are outputs from program lorenz1EH!!!
  //
  ::strcat(sysCmd, " 1>/tmp/runLorenz1 2>/tmp/runLorenz2");
  if ((ofd != NULL) && (::logging > 0))
    ::fprintf(ofd, "lorenzModel() sysCmd=%s\n", sysCmd);
  //
  // Invoke script runLorenz to produce via gnuplot
  // a file called lorenz.jpg
  localRet = ::system(sysCmd);
  //
  if ((ofd != NULL) && (::logging > 0))
    ::fprintf(ofd, "lorenzModel() system(runLorenz) returns %d\n", localRet);
  //
  pTimeStr2 = getTime();;  // end of execution
  //
  // Prepare method replaceFields() in method send_page()
  // Construct the valueList containig the replacing field values
  // in exactly the same order as given in itemList
  pGlobalHttp->setSepChars('#', '|');
  ::strcpy(valueList, "");
  //
  ::strcat(valueList, feld1);   // inputs...
  ::strcat(valueList, "|");
  ::strcat(valueList, feld2);
  ::strcat(valueList, "|");
  ::strcat(valueList, feld3);
  ::strcat(valueList, "|");
  ::strcat(valueList, feld4);
  ::strcat(valueList, "|");
  // 
  localRet = ::readOutputFields(feld5, feld6, feld7, 64);
  if ((ofd != NULL) && (::logging > 0))
    ::fprintf(ofd, "lorenzModel() readOutFields() ret=%d\n", localRet);
  ::strcat(valueList, feld5);   // outputs
  ::strcat(valueList, "|");
  ::strcat(valueList, feld6);
  ::strcat(valueList, "|");
  ::strcat(valueList, feld7);
  ::strcat(valueList, "|");
  ::strcat(valueList, result);
  ::strcat(valueList, "|");
  //
  // add standard fields date1, date2 and time
  pDateStr = getDateYYYYMMDD();
  if (pDateStr != NULL)
  {
    ::strcat(valueList, pDateStr);
    delete pDateStr;
  }  
  ::strcat(valueList, "|");
  //
  if (pTimeStr1 != NULL)
  {
    ::strcat(valueList, pTimeStr1);
    delete pTimeStr1;
  }
  ::strcat(valueList, "|");
  if (pTimeStr2 != NULL)
  {
    ::strcat(valueList, pTimeStr2);
    delete pTimeStr2;
  }
  //
  // add the runID  to the fields
  ::strcat(valueList, "|");
  runId = getNextRunID();
  ::sprintf(xBuf, "%ld", runId);
  ::strcat(valueList, xBuf);
  ::strcat(valueList, "|");
  //
  /*
  // add the URL for the PDF document to the fields
  ::strcpy(xBuf, host_name);
  ::strcat(xBuf, "/");
  caddr = ::getBasename(sineFileName);
  ::strcat(xBuf, caddr);
  delete caddr;
  ::strcat(xBuf, ".pdf");
  ::strcat(valueList, xBuf);
  ::strcat(valueList, "|");
  */
  //
  // The answer page will now be sent out after replacing the
  // item fields by the corresponging value fields 
  iRet = pGlobalHttp->send_page(templatePageFn,   // answerLorenz.htm
                                itemList,
                                valueList);
  //
  zurueck:
  if ((ofd != NULL) && (::logging > 0))
  {
    ::fprintf(ofd, "lorenzModel() end ret=%d\n", iRet);
  }  
  return iRet;
}   // end lorenzModel()

// -----------------------------------------------------------------------
void printHex(FILE *ofd,
              const char *pBuf, int lBuf)
{
  for (int i=0; i < lBuf; i++)
  {
    ::fprintf(ofd, "%02x ", pBuf[i] & 0xff);
  }
  ::fprintf(ofd, "\n");
}   // end printHex()

// ------------------------------------------------------------------------
// actually not used!!!
// metereological function for J.Steppeler
// alternativ answer
int function4()
{
  int iRet = 0;
  int locRet = 0;
  int ifd = -1;
  char buf[256];
  struct stat statbuf;
  const char *pFileName = "/var/www/html/sineValues.pdf"; 
  //
  // funktioniert nicht!
  /*
  locRet = ::stat(pFileName, &statbuf);
  if (locRet < 0)
  {
    ::sprintf(buf, "ERROR: cannot stat() file=%s\n", pFileName);
    pGlobalHttp->send_error_document(::pProjectTitle,
                                     buf);
    iRet = -1;
    goto zurueck;
  }
  //
  ifd = ::open(pFileName, O_RDONLY);
  if (ifd < 0)
  {
    ::sprintf(buf, "ERROR: cannot open file=%s\n", pFileName);
    pGlobalHttp->send_error_document(::pProjectTitle,
                                     buf);
    iRet = -2;
    goto zurueck;
  }
  if (ofd != NULL)
    ::fprintf(ofd, "\n--- start of generated PDF file in HTML format ---\n");
  pGlobalHttp->send_header("Content-Type:", "application/pdf");
  if (ofd != NULL)
    ::fprintf(ofd, "Content-Type: application/pdf\n");
  ::sprintf(buf, "%lu", (unsigned long)(statbuf.st_size));
  pGlobalHttp->send_header("Content-Length:", buf); 
  ::printf("\n");  // headers must be terminated by an empty line
            (unsigned long)(statbuf.st_size));
  if (ofd != NULL)
    ::fprintf(ofd, "Content-Length: %lu\n\n", 
                (unsigned long)(statbuf.st_size));
  ::printf("<HTML>\n");
  if (ofd != NULL)
    ::fprintf(ofd, "<HTML>\n");
  ::printf("<BODY>\n");
  if (ofd != NULL)
    ::fprintf(ofd, "<BODY>\n");
  //
  while (1)
  {
    locRet = ::read(ifd, buf, sizeof(buf));
    if (locRet <= 0)
      break;
    ::write(1, buf, locRet);
    if (ofd != NULL)
      printHex(ofd, buf, locRet);
  }
  ::printf("</BODY>\n");
  if (ofd != NULL)
    ::fprintf(ofd, "</BODY>\n");
  ::printf("</HTML>\n");
  if (ofd != NULL)
    ::fprintf(ofd, "</HTML>\n");
  //
  //
  zurueck:
  if (ifd >= 0)
    ::close(ifd);
  */
  // use server redirection
  // Das funktioniert und ist sehr einfach!
  pGlobalHttp->send_header("Location:", "/sineValues.pdf");
  ::printf("\n");  // headers must be terminated by an empty line^

  return iRet;
}   // end function4()

// ------------------------------------------------------------------------
// main CGI program
int main(int argc, char *argv[])
{
  int  handled, n, ll, ret;
  int  localRet = 0;
  long no_satz;
  char *caddr;
  char xbuf[128];
  char *sDate = NULL;
  char *sTime = NULL;
  //
  // Protokolldatei im Modus "Append" oeffnen
  ofd = ::fopen(proto_fn, "wa");
  if (ofd == NULL)
  {
    ::proto_error = -1;
    send_error_document(::pProjectTitle,
                         "Could not open protocol file");
    ret = -1;
    goto zurueck;
  }
  else
  {
    ret = ::fprintf(ofd, "Protokolldatei ist OK\n");
    ::proto_error = ret;
  }
  pGlobalHttp = new HU_HTTP();
  sDate = get_german_date();
  sTime = get_time();
  if (ofd != NULL) 
    ::fprintf(ofd, "Datum: %s   Zeit: %s\n", sDate, sTime);
  //
  // read initial values from the given ini file
  pIni = new IniFile();
  pIni->setLogging(0);
  localRet = pIni->load(pIniFileName);
  if (localRet > 0)
  {
    if (ofd != NULL)
      ::fprintf(ofd, "%d entries in inifile=%s\n", localRet, pIniFileName);
  }
  else
  {
    if (ofd != NULL)
      ::fprintf(ofd, "ERROR: no ini file=%s\n", pIniFileName);
    send_error_document(::pProjectTitle,
                        "ERROR no ini file");
    ret = -2;
    goto zurueck;
  }
  // 
  // evaluate some entries from the ini file
  caddr = ::pIni->get("general", "template1");
  if ((caddr != NULL) && (*caddr != 0))
  {
    ::strcpy(templatePageFn, caddr);
  }
  else
    send_error_document(::pProjectTitle,
                        "ERROR: no entry template1 in ini file");
  //
  caddr = ::pIni->get("general", "logging");
  if ((caddr != NULL) && (*caddr != 0))
  {
    ::logging = ::atoi(caddr);  // level of logging!
    // Note: assumes that file proto_fn has been open
    // output of logging will be written via file descriptor ofd
  }
  if (ofd != NULL)
    ::fprintf(ofd, "template page=%s\n", templatePageFn);
  //
  // evaluate the essential environment variables
  get_env_vars();

  // get the query parameters
  // string ::parbuf will contain all customer-related fields from the form
  ::parbuf = get_query_params();
  if (ofd != NULL)
  {
    ::fprintf(ofd, "QUERY_STRING=%s\n", query_string);
    ::fprintf(ofd, "parbuf=%s\n", ::parbuf);
    ::fflush(ofd);
  }
  // 
  // Funktionsverteiler
  handled = 0;

  if (::strchr("0123456789", query_string[0]) != NULL)
  {
      if (ofd != NULL) 
        ::fprintf(ofd, "starting lorenzModel()\n");
     // calculate plot ploints following the Lorenz model
     // script runLorenz invoked in lorenzModel contains the logic
     // A JPG file called lorenz.jpg will be produced
     lorenzModel();
     handled = 1;
  }
  if (handled == 0)
  {
    char buf[LBUF];
    ::strcpy(buf, "ERROR: URL nicht erkannt, query_string=");
    ::strcat(buf, query_string);
    ::strcat(buf, "<BR>\n");
    ::strcat(buf, 
           "CGI address should be: addr:80/cgi-bin/metereo2/?1 or similar");
    ::strcat(buf, "<BR>\n");
    send_error_document(::pProjectTitle,
                        buf);
  }
  //
  // Aufraeumarbeiten
  if (::parbuf != NULL)
  {
    delete ::parbuf;
    ::parbuf = NULL;
  }
  //
  zurueck:
  //
  // Protokolldatei schliessen
  if (ofd != NULL) 
  {
    ::fprintf(ofd, "----------------------------------------------------\n");
    ::fflush(ofd);
    ::fclose(ofd);
  } 
  if (pGlobalHttp != NULL)
    delete pGlobalHttp;
  if (pIni != NULL)
    delete pIni;
  return 0;
}   /* end main */
